from __future__ import absolute_import, unicode_literals

from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from deux.authtoken import views

urlpatterns = [
    url(r"^login/$", views.ObtainMFAAuthToken.as_view(),
        name="login"),
]

urlpatterns = format_suffix_patterns(urlpatterns)



from __future__ import absolute_import, unicode_literals

import six
from mock import patch

from django.core.urlresolvers import reverse
from rest_framework import status

from deux.app_settings import mfa_settings
from deux.constants import SMS
from deux.tests.test_base import BaseUserTestCase


class _BaseMFAViewTest(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa_1 = mfa_settings.MFA_MODEL.objects.create(user=self.user1)
        self.mfa_2 = mfa_settings.MFA_MODEL.objects.create(user=self.user2)
        self.mfa_2.enable(SMS)
        self.phone_number = "1234567890"
        self.mfa_2.phone_number = self.phone_number
        self.mfa_2.save()


class ObtainMFAAuthTokenTest(_BaseMFAViewTest):
    url = reverse("authtoken:login")

    @patch("deux.authtoken.serializers.MultiFactorChallenge")
    def test_login_mfa_required(self, multifactorchallenge):
        # Correct credentials without MFA code.
        data = {
            "username": self.user2.username,
            "password": self.password2
        }
        resp = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data)
        self.assertEqual(resp.data, {
            "mfa_type": SMS,
            "mfa_required": True
        })
        with self.assertRaises(KeyError):
            resp.data["token"]

    def test_login_mfa_not_required(self):
        # Incorrect password.
        data = {"username": self.user1.username,
                "password": "incorrect_password"}
        resp = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data)
        self.assertEqual(resp.data, {
            "non_field_errors": [
                "Unable to log in with provided credentials."
            ]
        })

        # Correct password.
        data["password"] = self.password1
        resp = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data)
        self.assertEqual(resp.data, {
            "token": six.text_type(self.user1.auth_token),
        })



from __future__ import absolute_import, unicode_literals

from deux.app_settings import mfa_settings
from deux.constants import DISABLED, SMS

from .test_base import BaseUserTestCase


class MultiFactorAuthTests(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa = mfa_settings.MFA_MODEL.objects.create(
            user=self.user1, phone_number="12345678900")

    def test_default_disabled(self):
        self.assertEqual(self.mfa.challenge_type, DISABLED)
        self.assertFalse(self.mfa.enabled)
        self.assertEqual(self.mfa.backup_code, "")

    def test_set_sms_enabled(self):
        self.mfa.enable(SMS)
        self.assertEqual(self.mfa.challenge_type, SMS)
        self.assertTrue(self.mfa.enabled)
        self.assertEqual(
            len(self.mfa.backup_code), mfa_settings.BACKUP_CODE_DIGITS)

    def test_disable(self):
        self.mfa.enable(SMS)
        self.assertEqual(self.mfa.challenge_type, SMS)

        self.mfa.disable()
        self.assertEqual(self.mfa.challenge_type, DISABLED)
        self.assertFalse(self.mfa.enabled)

    def test_phone_number(self):
        self.assertEqual(self.mfa.phone_number, "12345678900")

    def test_backup_code_generation(self):
        # AssertionError if disabled.
        with self.assertRaises(AssertionError):
            self.mfa.refresh_backup_code()

        # Valid backup code if enabled.
        self.mfa.enable(SMS)
        self.assertEqual(
            len(self.mfa.refresh_backup_code()),
            mfa_settings.BACKUP_CODE_DIGITS
        )

        # Codes should not be same after regeneration.
        current_code = self.mfa.backup_code
        new_code = self.mfa.refresh_backup_code()
        self.assertIsNotNone(new_code)
        self.assertNotEquals(current_code, new_code)

    def test_check_and_use_backup_code(self):
        self.mfa.enable(SMS)
        self.mfa.refresh_backup_code()

        # Test for using incorrect code.
        bad_code = "123456abcdef"
        self.assertFalse(self.mfa.check_and_use_backup_code(bad_code))

        # Test for using correct code.
        code = self.mfa.backup_code
        self.assertTrue(self.mfa.check_and_use_backup_code(code))
        instance = mfa_settings.MFA_MODEL.objects.get(user=self.user1)
        self.assertFalse(instance.enabled)
        self.assertEqual(instance.challenge_type, DISABLED)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import os

from sphinx_celery import conf

globals().update(conf.build_config(
    'deux', __file__,
    project='deux',
    canonical_url='http://deux.readthedocs.io',
    webdomain='robinhood.com',
    github_project='robinhood/deux',
    copyright='2016',
    html_logo='images/logo.png',
    html_favicon='images/favicon.ico',
    html_static_path=[],
    include_intersphinx={'python', 'sphinx'},
    django_settings='test_proj.settings',
    apicheck_package='deux',
    apicheck_ignore_modules=[
        'deux.authtoken.tests.*',
        'deux.authtoken.urls',
        'deux.migrations.*',
        'deux.locale.*',
        'deux.oauth2.tests.*',
        'deux.oauth2.urls',
        'deux.tests.*',
        'deux.urls',
        'deux.app_settings',
    ],
    spelling_word_list_filename='spelling/spelling_wordlist.txt',
))

html_theme_path = ['theme']
html_theme = 'deux'

def configcheck_project_settings():
    return set()



from __future__ import absolute_import, unicode_literals

import json

from oauth2_provider.oauth2_backends import JSONOAuthLibCore


class MFARequestBackend(JSONOAuthLibCore):
    """
    class::MFARequestBackend()

    OAuth2 backend class for MFA extending ``JSONOAuthLibCore``. It extracts
    extra credentials (``mfa_code`` and ``backup_code``) from the request body.
    """

    def _get_extra_credentials(self, request):
        """
        Gets dictionary of ``mfa_code`` and ``backup_code`` from the request.

        :param request: The API request to this MFA backend class.
        :returns: Dictionary with ``mfa_code`` and ``backup_code``.
        """
        body = json.loads(request.body.decode("utf-8"))
        return {
            "mfa_code": body.get("mfa_code"),
            "backup_code": body.get("backup_code"),
        }



from __future__ import absolute_import, unicode_literals

from django.utils.translation import ugettext_lazy as _

#: Error if user submits both MFA and backup code for authentication.
BOTH_CODES_ERROR = _(
    "Login does not take both a verification and backup code.")

#: Error if MFA is unexpectedly in a disabled state.
DISABLED_ERROR = _("Two factor authentication is not enabled.")

#: Error if MFA is unexpectedly in an enabled state.
ENABLED_ERROR = _("Two factor authentication is already enabled.")

#: Error if an invalid backup code is entered.
INVALID_BACKUP_CODE_ERROR = _("Please enter a valid backup code.")

#: Error if a user provides an invalid username/password combination.
INVALID_CREDENTIALS_ERROR = _("Unable to log in with provided credentials.")

#: Error if an invalid MFA code is entered.
INVALID_MFA_CODE_ERROR = _("Please enter a valid code.")

#: Error if an invalid phone number is entered.
INVALID_PHONE_NUMBER_ERROR = _("Please enter a valid phone number.")

#: Error if phone number is not set for a challenge that requires it.
PHONE_NUMBER_NOT_SET_ERROR = _(
    "MFA phone number must be set for this challenge.")

#: Error if SMS fails to send.
SMS_SEND_ERROR = _("SMS failed to send.")

#: Message body for a MFA code.
MFA_CODE_TEXT_MESSAGE = _("Two Factor Authentication Code: {code}")



from __future__ import absolute_import, unicode_literals

import importlib
import six

from django.conf import settings


USER_SETTINGS = getattr(settings, 'DEUX', None)

DEFAULTS = {
    "BACKUP_CODE_DIGITS": 12,
    "MFA_CODE_NUM_DIGITS": 6,
    "MFA_MODEL": "deux.models.MultiFactorAuth",
    "SEND_MFA_TEXT_FUNC": "deux.notifications.send_mfa_code_text_message",
    "STEP_SIZE": 30,
    "TWILIO_ACCOUNT_SID": "",
    "TWILIO_AUTH_TOKEN": "",
    "TWILIO_SMS_POOL_SID": "",
}

# List of settings that cannot be empty.
MANDATORY = ()

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'MFA_MODEL',
    'SEND_MFA_TEXT_FUNC',
)


def perform_import(val, setting_name):

    if isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):

    try:
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError:
        msg = "Coud not import {val} for setting {setting_name}".format(
            val=val, setting_name=setting_name)
        raise ImportError(msg)


class MFASettings(object):

    def __init__(self, user_settings=None, defaults=None, import_strings=None,
                 mandatory=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or {}
        self.import_strings = import_strings or ()
        self.mandatory = mandatory or ()

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid deux setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if val and attr in self.import_strings:
            val = perform_import(val, attr)

        self.validate_setting(attr, val)

        # Cache the result
        setattr(self, attr, val)
        return val

    def validate_setting(self, attr, val):
        if not val and attr in self.mandatory:
            raise AttributeError("deux setting: '%s' is mandatory" % attr)


mfa_settings = MFASettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS, MANDATORY)



from __future__ import absolute_import, unicode_literals

from rest_framework import generics
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import IsAuthenticated

from deux import strings
from deux.app_settings import mfa_settings
from deux.constants import SMS
from deux.serializers import (
    BackupCodeSerializer,
    MultiFactorAuthSerializer,
    SMSChallengeRequestSerializer,
    SMSChallengeVerifySerializer,
)


class MultiFactorAuthMixin(object):
    """
    class::MultiFactorAuthMixin()

    Mixin that defines queries for MFA objects.
    """

    def get_object(self):
        """Gets the current user's MFA instance"""
        instance, created = mfa_settings.MFA_MODEL.objects.get_or_create(
            user=self.request.user)
        return instance


class MultiFactorAuthDetail(
        MultiFactorAuthMixin, generics.RetrieveDestroyAPIView):
    """
    class::MultiFactorAuthDetail()

    View for requesting data about MultiFactorAuth and disabling MFA.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = MultiFactorAuthSerializer

    def perform_destroy(self, instance):
        """
        The delete method should disable MFA for this user.

        :raises rest_framework.exceptions.ValidationError: If MFA is not
            enabled.
        """
        if not instance.enabled:
            raise ValidationError({
                "detail": strings.DISABLED_ERROR
            })
        instance.disable()


class _BaseChallengeView(MultiFactorAuthMixin, generics.UpdateAPIView):
    """
    class::_BaseChallengeView()

    Base view for different challenges.
    """
    permission_classes = (IsAuthenticated,)

    @property
    def challenge_type(self):
        """
        Represents the challenge type this serializer represents.

        :raises NotImplemented: If the extending class does not define
            ``challenge_type``.
        """
        raise NotImplemented  # pragma: no cover


class SMSChallengeRequestDetail(_BaseChallengeView):
    """
    class::SMSChallengeRequestDetail()

    View for requesting SMS challenges to enable MFA through SMS.
    """
    challenge_type = SMS
    serializer_class = SMSChallengeRequestSerializer


class SMSChallengeVerifyDetail(_BaseChallengeView):
    """
    class::SMSChallengeVerifyDetail()

    View for verify SMS challenges to enable MFA through SMS.
    """
    challenge_type = SMS
    serializer_class = SMSChallengeVerifySerializer


class BackupCodeDetail(MultiFactorAuthMixin, generics.RetrieveAPIView):
    """
    class::BackupCodeDetail()

    View for retrieving the user's backup code.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = BackupCodeSerializer



# -*- coding: utf-8 -*-
"""Multifactor Authentication for Django Rest Framework"""
# :copyright: (c) 2016, Robinhood Markets.
#             All rights reserved.
# :license:   BSD (3 Clause), see LICENSE for more details.

from __future__ import absolute_import, unicode_literals

from collections import namedtuple

version_info_t = namedtuple(
    'version_info_t', ('major', 'minor', 'micro', 'releaselevel', 'serial'),
)

VERSION = version_info = version_info_t(1, 1, 1, '', '')

__version__ = '{0.major}.{0.minor}.{0.micro}{0.releaselevel}'.format(VERSION)
__author__ = 'Robinhood Markets'
__contact__ = 'opensource@robinhood.com'
__homepage__ = 'https://github.com/robinhood/deux'
__docformat__ = 'restructuredtext'

# -eof meta-

__all__ = []



from __future__ import absolute_import, unicode_literals

from rest_framework.authtoken.views import ObtainAuthToken
from rest_framework.authtoken.models import Token
from rest_framework.response import Response

from deux.authtoken.serializers import MFAAuthTokenSerializer


class ObtainMFAAuthToken(ObtainAuthToken):
    """
    class::ObtainMFAAuthToken()

    View for authenticating which extends the ``ObtainAuthToken`` from
    Django Rest Framework's Token Authentication.
    """
    serializer_class = MFAAuthTokenSerializer

    def post(self, request, *args, **kwargs):
        """
        function::post(self, request)

        Override ObtainAuthToken's post method for multifactor
        authentication.

        (1) When MFA is required, send the user a response
        indicating which challenge is required.
        (2) When authentication is successful return the auth token.

        :param request: Request object from the client.
        """
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data
        if "mfa_required" in data and data["mfa_required"]:
            return Response({
                "mfa_required": True,
                "mfa_type": serializer.validated_data["mfa_type"]
            })
        else:
            user = serializer.validated_data['user']
            token, created = Token.objects.get_or_create(user=user)
            return Response({"token": token.key})



from __future__ import absolute_import, unicode_literals

import six
from uuid import uuid4

from django.utils.crypto import constant_time_compare
from django_otp.oath import totp

from deux.app_settings import mfa_settings
from deux.constants import CHALLENGE_TYPES, SMS


def generate_mfa_code(bin_key, drift=0):
    """
    Generates an MFA code based on the ``bin_key`` for the current timestamp
    offset by the ``drift``.

    :param bin_key: The secret key to be converted into an MFA code
    :param drift: Number of time steps to shift the conversion.
    """
    return six.text_type(totp(
        bin_key,
        step=mfa_settings.STEP_SIZE,
        digits=mfa_settings.MFA_CODE_NUM_DIGITS,
        drift=drift
    )).zfill(mfa_settings.MFA_CODE_NUM_DIGITS)


def generate_key():
    """Generates a key used for secret keys."""
    return uuid4().hex


def verify_mfa_code(bin_key, mfa_code):
    """
    Verifies that the inputted ``mfa_code`` is a valid code for the given
    secret key. We check the ``mfa_code`` against the current time stamp as
    well as one time step before and after.

    :param bin_key: The secret key to verify the MFA code again.
    :param mfa_code: The code whose validity this function tests.
    """
    if not mfa_code:
        return False
    try:
        mfa_code = int(mfa_code)
    except ValueError:
        return False
    else:
        totp_check = lambda drift: int(
            generate_mfa_code(bin_key=bin_key, drift=drift))
        return any(
            constant_time_compare(totp_check(drift), mfa_code)
            for drift in [-1, 0, 1]
        )


class MultiFactorChallenge(object):
    """
    A class that represents a supported challenge and has the ability to
    execute the challenge.

    :param instance: :class:`MultiFactorAuth` instance to use.
    :param challenge_type: Challenge type being used for this object.
    :raises AssertionError: If ``challenge_type`` is not a supported
        challenge type.
    """

    def __init__(self, instance, challenge_type):
        assert challenge_type in CHALLENGE_TYPES, (
            "Inputted challenge type is not supported."
        )
        self.instance = instance
        self.challenge_type = challenge_type

    def generate_challenge(self):
        """
        Generates and executes the challenge object based on the challenge
        type of this object.
        """
        dispatch = {
            SMS: self._sms_challenge
        }
        for challenge in CHALLENGE_TYPES:
            assert challenge in dispatch, (
                "'{challenge}' does not have a challenge dispatch "
                "method.".format(challenge=challenge)
            )
        return dispatch[self.challenge_type]()

    def _sms_challenge(self):
        """Executes the SMS challenge."""
        code = generate_mfa_code(bin_key=self.instance.sms_bin_key)
        mfa_settings.SEND_MFA_TEXT_FUNC(
            mfa_instance=self.instance, mfa_code=code)



from __future__ import absolute_import, unicode_literals

from django.contrib.auth.models import User
from rest_framework.test import APITestCase


class BaseUserTestCase(APITestCase):

    def simpleUserSetup(self):
        self.password1 = "password_1"
        self.user1 = User.objects.create_user(
            username="test1",
            email="test1@example.com",
            password=self.password1,
            first_name="Test",
            last_name="One")

        self.password2 = "password_2"
        self.user2 = User.objects.create_user(
            username="test2",
            email="test2@example.com",
            password=self.password2,
            first_name="Test",
            last_name="Two")

    def check_get_response(
            self, url, expected_status_code, data=None, user=None):
        self.client.force_authenticate(user)
        response = self.client.get(url, data=data)
        self.assertEqual(response.status_code, expected_status_code)
        return response

    def check_post_response(
            self, url, expected_status_code, data=None, user=None,
            format="json", headers=None):
        headers = headers or {}
        self.client.force_authenticate(user)
        response = self.client.post(url, data=data, format=format, **headers)
        self.assertEqual(response.status_code, expected_status_code)
        return response

    def check_put_response(
            self, url, expected_status_code, data=None, user=None,
            format="json"):
        self.client.force_authenticate(user)
        response = self.client.put(url, data=data, format=format)
        self.assertEqual(response.status_code, expected_status_code)
        return response

    def check_patch_response(
            self, url, expected_status_code, data=None, user=None,
            format="json"):
        self.client.force_authenticate(user)
        response = self.client.patch(url, data=data, format=format)
        self.assertEqual(response.status_code, expected_status_code)
        return response

    def check_delete_response(
            self, url, expected_status_code, user=None):
        self.client.force_authenticate(user)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, expected_status_code)
        return response

    def check_options_response(
            self, url, expected_status_code, data=None, user=None,
            format="json"):
        self.client.force_authenticate(user)
        response = self.client.options(url, data=data, format=format)
        self.assertEqual(response.status_code, expected_status_code)
        return response



from __future__ import absolute_import, unicode_literals

import six
from mock import patch

from deux.app_settings import mfa_settings
from deux.constants import DISABLED, SMS
from deux.authtoken.serializers import MFAAuthTokenSerializer
from deux.services import generate_mfa_code
from deux.tests.test_base import BaseUserTestCase


class MFAAuthTokenSerializerTest(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa = mfa_settings.MFA_MODEL.objects.create(user=self.user2)

    def test_login_with_no_mfa_object(self):
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user1.username,
            "password": "incorrect_password",
        })
        self.assertFalse(serializer.is_valid())

        serializer = MFAAuthTokenSerializer(data={
            "username": self.user1.username,
            "password": self.password1
        })
        self.assertTrue(serializer.is_valid())

    def test_login_with_disabled_mfa_object(self):
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": "incorrect_password",
        })
        self.assertFalse(serializer.is_valid())

        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2
        })
        self.assertTrue(serializer.is_valid())

    def test_login_fail_with_both_codes(self):
        self.mfa.enable(SMS)
        self.mfa.refresh_backup_code()

        mfa_code = generate_mfa_code(self.mfa.sms_bin_key)
        backup_code = self.mfa.backup_code

        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2,
            "mfa_code": mfa_code,
            "backup_code": backup_code
        })
        self.assertFalse(serializer.is_valid())

    def test_login_with_mfa_code(self):
        self.mfa.enable(SMS)
        mfa_code = generate_mfa_code(self.mfa.sms_bin_key)
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2,
            "mfa_code": mfa_code
        })
        self.assertTrue(serializer.is_valid())

        bad_code = six.text_type(int(mfa_code) + 1)
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2,
            "mfa_code": bad_code
        })
        self.assertFalse(serializer.is_valid())

    def test_login_with_backup_code(self):
        self.mfa.enable(SMS)
        bad_code = "abcdef123456"
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2,
            "backup_code": bad_code
        })
        self.assertFalse(serializer.is_valid())

        backup_code = self.mfa.backup_code
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2,
            "backup_code": backup_code
        })
        self.assertTrue(serializer.is_valid(raise_exception=True))
        instance = mfa_settings.MFA_MODEL.objects.get(user=self.user2)
        self.assertFalse(instance.enabled)
        self.assertEqual(instance.challenge_type, DISABLED)
        self.assertEqual(instance.backup_code, "")

    @patch("deux.authtoken.serializers.MultiFactorChallenge")
    def test_login_and_continue_with_challenge(self, challenge):
        self.mfa.enable(SMS)
        self.mfa.phone_number = "1234567890"
        self.mfa.save()
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": "incorrect_password",
        })
        self.assertFalse(serializer.is_valid())

        # With correct username and password, response should require MFA.
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user2.username,
            "password": self.password2
        })
        self.assertTrue(serializer.is_valid())
        data = serializer.validated_data
        self.assertTrue(data["mfa_required"])
        self.assertEqual(data["mfa_type"], SMS)
        challenge.assert_called_once_with(self.mfa, SMS)
        challenge.return_value.generate_challenge.assert_called_once_with()

    def test_login_with_other_users_code(self):
        mfa_1 = mfa_settings.MFA_MODEL.objects.create(user=self.user1)
        mfa_2 = self.mfa

        mfa_1.enable(SMS)
        mfa_2.enable(SMS)

        # User 1 using User 2's MFA Code should fail.
        mfa_2_code = generate_mfa_code(mfa_2.sms_bin_key)
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user1.username,
            "password": self.password1,
            "mfa_code": mfa_2_code
        })
        self.assertFalse(serializer.is_valid())

        # User 1 using User 2's Backup Code should fail.
        mfa_2_backup = mfa_2.backup_code
        serializer = MFAAuthTokenSerializer(data={
            "username": self.user1.username,
            "password": self.password1,
            "backup_code": mfa_2_backup
        })
        self.assertFalse(serializer.is_valid())



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-22 21:31
from __future__ import absolute_import, unicode_literals

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import deux.services


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0007_alter_validators_add_error_messages'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultiFactorAuth',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='multi_factor_auth', serialize=False, to=settings.AUTH_USER_MODEL)),
                ('phone_number', models.CharField(blank=True, default='', max_length=15, validators=[django.core.validators.RegexValidator(message='Please enter a valid phone number.', regex='^(\\d{7,15})$')])),
                ('challenge_type', models.CharField(blank=True, choices=[('sms', 'SMS'), ('', 'Off')], default='', max_length=16)),
                ('backup_key', models.CharField(blank=True, default='', help_text='Hex-Encoded Secret Key', max_length=32)),
                ('sms_secret_key', models.CharField(default=deux.services.generate_key, help_text='Hex-Encoded Secret Key', max_length=32)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import absolute_import, unicode_literals

from oauth2_provider.oauth2_validators import OAuth2Validator

from django.contrib.auth import authenticate
from django.utils.encoding import force_text

from deux import strings
from deux.oauth2.exceptions import (
    ChallengeRequiredMessage,
    InvalidLoginError,
)
from deux.services import MultiFactorChallenge, verify_mfa_code


class MFAOAuth2Validator(OAuth2Validator):
    """
    class::MFAOAuth2Validator()

    OAuth2 validator class for MFA that validates requests to authenticate
    with username and password by also verifying that they supply the correct
    MFA code or backup code if multifactor authentication is enabled.
    """

    def validate_user(
            self, username, password, client, request, *args, **kwargs):
        """
        Overrides the OAuth2Validator validate method to implement multi factor
        authentication.

        If MFA is disabled, authentication requires just a username and
        password.

        If MFA is enabled, authentication requires a username, password,
        and either a MFA code or a backup code. If the request only provides
        the username and password, the server will generate an appropriate
        challenge and respond with `mfa_required = True`.

        Upon using a backup code to authenticate, MFA will be disabled.

        :param attrs: Dictionary of data inputted by the user.
        :raises deux.oauth2.exceptions.InvalidLoginError: If invalid MFA
            code or backup code are submitted. Also if both types of code are
            submitted simultaneously.
        :raises deux.oauth2.exceptions.ChallengeRequiredMessage: If the user
            has MFA enabled but only supplies the correct username and
            password. This exception will prompt the OAuth2 system to send a
            response asking the user to supply an MFA code.
        """

        user = authenticate(username=username, password=password)
        if not (user and user.is_active):
            raise InvalidLoginError(force_text(
                strings.INVALID_CREDENTIALS_ERROR))

        mfa = None
        if hasattr(user, "multi_factor_auth"):
            mfa = user.multi_factor_auth

        if mfa and mfa.enabled:
            mfa_code = request.extra_credentials.get("mfa_code")
            backup_code = request.extra_credentials.get("backup_code")

            if mfa_code and backup_code:
                raise InvalidLoginError(force_text(strings.BOTH_CODES_ERROR))
            elif mfa_code:
                bin_key = mfa.get_bin_key(mfa.challenge_type)
                if not verify_mfa_code(bin_key, mfa_code):
                    raise InvalidLoginError(force_text(
                        strings.INVALID_MFA_CODE_ERROR))
            elif backup_code:
                if not mfa.check_and_use_backup_code(backup_code):
                    raise InvalidLoginError(force_text(
                        strings.INVALID_BACKUP_CODE_ERROR))
            else:
                challenge = MultiFactorChallenge(mfa, mfa.challenge_type)
                challenge.generate_challenge()
                raise ChallengeRequiredMessage(mfa.challenge_type)
        request.user = user
        return True



from __future__ import absolute_import, unicode_literals

from twilio.rest import TwilioRestClient
from twilio.rest.exceptions import TwilioRestException

from deux import strings
from deux.app_settings import mfa_settings
from deux.exceptions import InvalidPhoneNumberError, TwilioMessageError

#: Error code from Twilio to indicate at ``InvalidPhoneNumberError``
NOT_SMS_DEVICE_CODE = 21401


def send_mfa_code_text_message(mfa_instance, mfa_code):
    """
    Sends the MFA Code text message to the user.

    :param mfa_instance: :class:`MultiFactorAuth` instance to use.
    :param mfa_code: MFA code in the form of a string.

    :raises deux.exceptions.InvalidPhoneNumberError: To tell system that this
        MFA object's phone number if not a valid number to receive SMS's.
    :raises deux.exceptions.TwilioMessageError: To tell system that Twilio
        failed to send message.
    """

    sid = mfa_settings.TWILIO_ACCOUNT_SID
    token = mfa_settings.TWILIO_AUTH_TOKEN
    twilio_num = mfa_settings.TWILIO_SMS_POOL_SID
    if not sid or not token or not twilio_num:
        print("Please provide Twilio credentials to send text messages. For "
              "testing purposes, the MFA code is {code}".format(code=mfa_code))
        return

    twilio_client = TwilioRestClient(sid, token)
    try:
        twilio_client.messages.create(
            body=strings.MFA_CODE_TEXT_MESSAGE.format(code=mfa_code),
            to=mfa_instance.phone_number,
            from_=twilio_num
        )
    except TwilioRestException as e:
        if e.code == NOT_SMS_DEVICE_CODE:
            raise InvalidPhoneNumberError()
        raise TwilioMessageError()



from __future__ import absolute_import, unicode_literals

from deux import strings


class FailedChallengeError(Exception):
    """Generic exception for a failed challenge execution."""
    pass


class InvalidPhoneNumberError(FailedChallengeError):
    """
    Exception for SMS that fails because phone number is not a valid
    number for receiving SMS's.
    """

    def __init__(self, message=strings.INVALID_PHONE_NUMBER_ERROR):
        super(InvalidPhoneNumberError, self).__init__(message)


class TwilioMessageError(FailedChallengeError):
    """
    Exception that Twilio failed to send the text message for reasons
    other than ``NotSMSNumberError``.
    """

    def __init__(self, message=strings.SMS_SEND_ERROR):
        super(TwilioMessageError, self).__init__(message)



from __future__ import absolute_import, unicode_literals

import six

from rest_framework import serializers

from deux.app_settings import mfa_settings
from deux import strings
from deux.constants import SMS
from deux.exceptions import FailedChallengeError
from deux.services import MultiFactorChallenge, verify_mfa_code


class MultiFactorAuthSerializer(serializers.ModelSerializer):
    """
    class::MultiFactorAuthSerializer()

    Basic MultiFactorAuthSerializer that encodes MFA objects into a standard
    response.

    The standard response returns whether MFA is enabled, the challenge
    type, and the user's phone number.
    """

    def to_representation(self, mfa_instance):
        """
        Encodes an MFA instance as the standard response.

        :param mfa_instance: :class:`MultiFactorAuth` instance to use.
        :returns: Dictionary with ``enabled``, ``challengetype``, and
            ``phone_number`` from the MFA instance.
        """
        data = {"enabled": mfa_instance.enabled}
        if mfa_instance.enabled:
            data["challenge_type"] = mfa_instance.challenge_type
        if mfa_instance.phone_number:
            data["phone_number"] = mfa_instance.phone_number
        return data

    class Meta:
        model = mfa_settings.MFA_MODEL


class _BaseChallengeRequestSerializer(MultiFactorAuthSerializer):
    """
    class::_BaseChallengeRequestSerializer()

    Base Serializer class for all challenge request.
    """

    @property
    def challenge_type(self):
        """
        Represents the challenge type this serializer represents.

        :raises NotImplemented: If the extending class does not define
            ``challenge_type``.
        """
        raise NotImplemented  # pragma: no cover

    def execute_challenge(self, instance):
        """
        Execute challenge for this instance based on the ``challenge_type``.

        :param instance: :class:`MultiFactorAuth` instance to use.
        :raises serializers.ValidationError: If the challenge fails to execute.
        """
        try:
            MultiFactorChallenge(
                instance, self.challenge_type).generate_challenge()
        except FailedChallengeError as e:
            raise serializers.ValidationError({
                "detail": six.text_type(e)
            })

    def validate(self, internal_data):
        """
        Validate the request to enable MFA through this challenge.

        Extending classes should extend for additional functionality. The
        base functionality ensures that MFA is not already enabled.

        :param internal_data: Dictionary of the request data.
        :raises serializers.ValidationError: If MFA is already enabled.
        """
        if self.instance.enabled:
            raise serializers.ValidationError({
                "detail": strings.ENABLED_ERROR
            })
        return internal_data

    def update(self, mfa_instance, validated_data):
        """
        If the request is valid, the serializer calls update which executes
        the ``challenge_type``.

        :param mfa_instance: :class:`MultiFactorAuth` instance to use.
        :param validated_data: Data returned by ``validate``.
        """
        self.execute_challenge(mfa_instance)


class _BaseChallengeVerifySerializer(MultiFactorAuthSerializer):
    """
    class::_BaseChallengeVerifySerializer()

    This serializer first extracts MFA code from request body
    (`to_internal_value`). It then  verifies the code against the
    corresponding `MultiFactorAuth` instance (`validate`). If the code
    is valid, it enables MFA based on the challenge type (`update`).
    """

    #: Requests to verify an MFA code must include an ``mfa_code``.
    mfa_code = serializers.CharField()

    @property
    def challenge_type(self):
        """
        Represents the challenge type this serializer represents.

        :raises NotImplemented: If the extending class does not define
            ``challenge_type``.
        """
        raise NotImplemented  # pragma: no cover

    def validate(self, internal_data):
        """
        Validates the request to verify the MFA code. It first ensures that
        MFA is not already enabled and then verifies that the MFA code is the
        correct code.

        :param internal_data: Dictionary of the request data.
        :raises serializers.ValidationError: If MFA is already enabled or if
            the inputted MFA code is not valid.
        """
        if self.instance.enabled:
            raise serializers.ValidationError({
                "detail": strings.ENABLED_ERROR
            })

        mfa_code = internal_data.get("mfa_code")
        bin_key = self.instance.get_bin_key(self.challenge_type)
        if not verify_mfa_code(bin_key, mfa_code):
            raise serializers.ValidationError({
                "mfa_code": strings.INVALID_MFA_CODE_ERROR
            })
        return {"mfa_code": mfa_code}

    def update(self, mfa_instance, validated_data):
        """
        If the request is valid, the serializer enables MFA on this instance
        for this serializer's ``challenge_type``.

        :param mfa_instance: :class:`MultiFactorAuth` instance to use.
        :param validated_data: Data returned by ``validate``.
        """
        mfa_instance.enable(self.challenge_type)
        return mfa_instance

    class Meta(MultiFactorAuthSerializer.Meta):
        fields = ("mfa_code",)


class SMSChallengeRequestSerializer(_BaseChallengeRequestSerializer):
    """
    class::SMSChallengeRequestSerializer()

    Serializer that facilitates a request to enable MFA over SMS.
    """

    #: This serializer represents the ``SMS`` challenge type.
    challenge_type = SMS

    def update(self, mfa_instance, validated_data):
        """
        If the request data is valid, the serializer executes the challenge
        by calling the super method and also saves the phone number the user
        requested the SMS to.

        :param mfa_instance: :class:`MultiFactorAuth` instance to use.
        :param validated_data: Data returned by ``validate``.
        """
        mfa_instance.phone_number = validated_data["phone_number"]
        super(SMSChallengeRequestSerializer, self).update(
            mfa_instance, validated_data)
        mfa_instance.save()
        return mfa_instance

    class Meta(_BaseChallengeRequestSerializer.Meta):
        fields = ("phone_number",)
        extra_kwargs = {
            "phone_number": {
                "required": True,
            },
        }


class SMSChallengeVerifySerializer(_BaseChallengeVerifySerializer):
    """
    class::SMSChallengeVerifySerializer()

    Extension of ``_BaseChallengeVerifySerializer`` that implements
    challenge verification for the SMS challenge.
    """

    #: This serializer represents the ``SMS`` challenge type.
    challenge_type = SMS


class BackupCodeSerializer(serializers.ModelSerializer):
    """
    class::BackupCodeSerializer()

    Serializer for the user requesting backup codes.
    """

    #: Serializer field for the backup code.
    backup_code = serializers.SerializerMethodField()

    def get_backup_code(self, instance):
        """
        Method for retrieving the backup code. On every request, the backup
        code is refreshed.

        :param instance: :class:`MultiFactorAuth` instance to use.
        :raises serializers.ValidationError: If MFA is disabled.
        """
        if self.instance.enabled:
            return self.instance.refresh_backup_code()
        else:
            raise serializers.ValidationError({
                "backup_code": strings.DISABLED_ERROR
            })

    class Meta:
        model = mfa_settings.MFA_MODEL
        fields = ("backup_code",)



from __future__ import absolute_import, unicode_literals

from mock import Mock, patch
from twilio.rest.exceptions import TwilioRestException

from deux.app_settings import mfa_settings
from deux.exceptions import InvalidPhoneNumberError, TwilioMessageError
from deux.notifications import send_mfa_code_text_message

from .test_base import BaseUserTestCase


class SendMFACodeTextMessageTests(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa = mfa_settings.MFA_MODEL.objects.create(user=self.user1)
        self.mfa.phone_number = "1234567890"
        self.mfa.save()
        self.code = "123456"

    @patch("deux.notifications.TwilioRestClient")
    @patch("deux.notifications.mfa_settings")
    def test_success(self, mfa_settings, twilio_client):
        mfa_settings.TWILIO_ACCOUNT_SID = "sid"
        mfa_settings.TWILIO_AUTH_TOKEN = "authtoken"
        mfa_settings.TWILIO_SMS_POOL_SID = "0987654321"

        twilio_client_instance = Mock()
        twilio_client.return_value = twilio_client_instance
        send_mfa_code_text_message(mfa_instance=self.mfa, mfa_code=self.code)
        twilio_client_instance.messages.create.assert_called_once_with(
            body="Two Factor Authentication Code: 123456",
            to="1234567890",
            from_="0987654321"
        )

    @patch("deux.notifications.TwilioRestClient")
    @patch("deux.notifications.mfa_settings")
    def test_invalid_number(self, mfa_settings, twilio_client):
        mfa_settings.TWILIO_ACCOUNT_SID = "sid"
        mfa_settings.TWILIO_AUTH_TOKEN = "authtoken"
        mfa_settings.TWILIO_SMS_POOL_SID = "0987654321"

        twilio_client_instance = Mock()
        twilio_client_instance.messages.create.side_effect = (
            TwilioRestException(400, "abc", code=21401))
        twilio_client.return_value = twilio_client_instance

        with self.assertRaises(InvalidPhoneNumberError):
            send_mfa_code_text_message(
                mfa_instance=self.mfa, mfa_code=self.code)

    @patch("deux.notifications.TwilioRestClient")
    @patch("deux.notifications.mfa_settings")
    def test_failed_sms_error(self, mfa_settings, twilio_client):
        mfa_settings.TWILIO_ACCOUNT_SID = "sid"
        mfa_settings.TWILIO_AUTH_TOKEN = "authtoken"
        mfa_settings.TWILIO_SMS_POOL_SID = "0987654321"

        twilio_client_instance = Mock()
        twilio_client_instance.messages.create.side_effect = (
            TwilioRestException(400, "abc"))
        twilio_client.return_value = twilio_client_instance

        with self.assertRaises(TwilioMessageError):
            send_mfa_code_text_message(
                mfa_instance=self.mfa, mfa_code=self.code)

    @patch("deux.notifications.TwilioRestClient")
    def test_no_twilio_credentials(self, twilio_client):
        twilio_client_instance = Mock()
        twilio_client.return_value = twilio_client_instance
        send_mfa_code_text_message(mfa_instance=self.mfa, mfa_code=self.code)
        twilio_client_instance.messages.create.assert_not_called()



from __future__ import absolute_import, unicode_literals

from django.core.exceptions import ValidationError
from django.test import TestCase

from deux.validators import phone_number_validator


class PhoneNumberValidatorTest(TestCase):

    def test_success(self):
        number = "123456789012345"
        success_tests = (number[:i] for i in range(7, 16))
        for phone_number in success_tests:
            phone_number_validator(phone_number)

    def test_fail(self):
        fail_tests = (
            None,
            "1",
            "123456",
            "123-321-1234",
            "123-123-1234x4321",
            "asdfghjkl;",
            "",
            "123456",
            "12345678901234567"
        )
        for phone_number in fail_tests:
            with self.assertRaises(ValidationError):
                phone_number_validator(phone_number)



from __future__ import absolute_import, unicode_literals

from django.utils.encoding import force_text
from rest_framework import serializers
from rest_framework.authtoken.serializers import AuthTokenSerializer

from deux import strings
from deux.services import MultiFactorChallenge, verify_mfa_code


class MFAAuthTokenSerializer(AuthTokenSerializer):
    """
    class::MFAAuthTokenSerializer()

    This extends the ``AuthTokenSerializer`` to support multifactor
    authentication.
    """

    #: Serializer field for MFA code field.
    mfa_code = serializers.CharField(required=False)

    #: Serializer field for Backup code.
    backup_code = serializers.CharField(required=False)

    def validate(self, attrs):
        """
        Extends the AuthTokenSerializer validate method to implement multi
        factor authentication.

        If MFA is disabled, authentication requires just a username and
        password.

        If MFA is enabled, authentication requires a username, password,
        and either a MFA code or a backup code. If the request only provides
        the username and password, the server will generate an appropriate
        challenge and respond with `mfa_required = True`.

        Upon using a backup code to authenticate, MFA will be disabled.

        :param attrs: Dictionary of data inputted by the user.
        :raises serializers.ValidationError: If invalid MFA code or backup code
            are submitted. Also if both types of code are submitted
            simultaneously.
        """
        attrs = super(MFAAuthTokenSerializer, self).validate(attrs)
        # User must exist if super method didn't throw error.
        user = attrs["user"]
        assert user is not None, "User should exist after super call."

        mfa = getattr(user, "multi_factor_auth", None)

        if mfa and mfa.enabled:
            mfa_code = attrs.get("mfa_code")
            backup_code = attrs.get("backup_code")

            if mfa_code and backup_code:
                raise serializers.ValidationError(
                    force_text(strings.BOTH_CODES_ERROR))
            elif mfa_code:
                bin_key = mfa.get_bin_key(mfa.challenge_type)
                if not verify_mfa_code(bin_key, mfa_code):
                    raise serializers.ValidationError(
                        force_text(strings.INVALID_MFA_CODE_ERROR))
            elif backup_code:
                if not mfa.check_and_use_backup_code(backup_code):
                    raise serializers.ValidationError(
                        force_text(strings.INVALID_BACKUP_CODE_ERROR))
            else:
                challenge = MultiFactorChallenge(mfa, mfa.challenge_type)
                challenge.generate_challenge()
                attrs["mfa_required"] = True
                attrs["mfa_type"] = mfa.challenge_type
        return attrs



from __future__ import absolute_import, unicode_literals

from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from deux.oauth2 import views

urlpatterns = [
    url(r'^token/$', views.MFATokenView.as_view(), name="token"),
]

urlpatterns = format_suffix_patterns(urlpatterns)



from __future__ import absolute_import, unicode_literals

from binascii import unhexlify

from django.contrib.auth.models import User
from django.db import models
from django.utils.crypto import constant_time_compare

from deux.app_settings import mfa_settings
from deux.constants import CHALLENGE_TYPES, DISABLED, SMS
from deux.services import generate_key
from deux.validators import phone_number_validator


class AbstractMultiFactorAuth(models.Model):
    """
    class::AbstractMultiFactorAuth()

    This abstract class holds user information, MFA status, and secret
    keys for the user.
    """

    #: Different status options for this MFA object.
    CHALLENGE_CHOICES = (
        (SMS, "SMS"),
        (DISABLED, "Off"),
    )

    #: User this MFA object represents.
    user = models.OneToOneField(
        User, related_name="multi_factor_auth", primary_key=True)

    #: User's phone number.
    phone_number = models.CharField(
        max_length=15, default="", blank=True,
        validators=[phone_number_validator])

    #: Challenge type used for MFA.
    challenge_type = models.CharField(
        max_length=16, default=DISABLED,
        blank=True, choices=CHALLENGE_CHOICES
    )

    #: Secret key used for backup code.
    backup_key = models.CharField(
        max_length=32, default="", blank=True,
        help_text="Hex-Encoded Secret Key"
    )

    #: Secret key used for SMS codes.
    sms_secret_key = models.CharField(
        max_length=32, default=generate_key,
        help_text="Hex-Encoded Secret Key"
    )

    @property
    def sms_bin_key(self):
        """Returns binary data of the SMS secret key."""
        return unhexlify(self.sms_secret_key)

    @property
    def enabled(self):
        """Returns if MFA is enabled."""
        return self.challenge_type in CHALLENGE_TYPES

    @property
    def backup_code(self):
        """Returns the users backup code."""
        return self.backup_key.upper()[:mfa_settings.BACKUP_CODE_DIGITS]

    def get_bin_key(self, challenge_type):
        """
        Returns the key associated with the inputted challenge type.

        :param challenge_type: The challenge type the key is requested for.
                               The type must be in the supported
                               `CHALLENGE_TYPES`.
        :raises AssertionError: If ``challenge_type`` is not a supported
                                challenge type.
        """
        assert challenge_type in CHALLENGE_TYPES, (
            "'{challenge}' is not a valid challenge type.".format(
                challenge=challenge_type)
        )
        return {
            SMS: self.sms_bin_key
        }.get(challenge_type, None)

    def enable(self, challenge_type):
        """
        Enables MFA for this user with the inputted challenge type.

        The enabling process includes setting this objects challenge type and
        generating a new backup key.

        :param challenge_type: Enable MFA for this type of challenge. The type
                               must be in the supported `CHALLENGE_TYPES`.
        :raises AssertionError: If ``challenge_type`` is not a supported
                                challenge type.
        """
        assert challenge_type in CHALLENGE_TYPES, (
            "'{challenge}' is not a valid challenge type.".format(
                challenge=challenge_type)
        )
        self.challenge_type = challenge_type
        self.backup_key = generate_key()
        self.save()

    def disable(self):
        """
        Disables MFA for this user.

        The disabling process includes setting the objects challenge type to
        `DISABLED`, and removing the `backup_key` and `phone_number`.
        """
        self.challenge_type = DISABLED
        self.backup_key = ""
        self.phone_number = ""
        self.save()

    def refresh_backup_code(self):
        """
        Refreshes the users backup key and returns a new backup code.

        This method should be used to request new backup codes for the user.
        """
        assert self.enabled, (
            "MFA must be on to run refresh_backup_codes."
        )
        self.backup_key = generate_key()
        self.save()
        return self.backup_code

    def check_and_use_backup_code(self, code):
        """
        Checks if the inputted backup code is correct and disables MFA if
        the code is correct.

        This method should be used for authenticating with a backup code. Using
        a backup code to authenticate disables MFA as a side effect.
        """
        backup = self.backup_code
        if code and constant_time_compare(code, backup):
            self.disable()
            return True
        return False

    class Meta:
        abstract = True



from __future__ import absolute_import, unicode_literals

from mock import patch

from django.core.urlresolvers import reverse
from rest_framework import status

from deux.app_settings import mfa_settings
from deux.constants import DISABLED, SMS
from deux.exceptions import FailedChallengeError
from deux.services import generate_mfa_code
from deux import strings

from .test_base import BaseUserTestCase


class _BaseMFAViewTest(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa_1 = mfa_settings.MFA_MODEL.objects.create(user=self.user1)
        self.mfa_2 = mfa_settings.MFA_MODEL.objects.create(user=self.user2)
        self.mfa_2.enable(SMS)
        self.phone_number = "1234567890"
        self.mfa_2.phone_number = self.phone_number
        self.mfa_2.save()


class MultiFactorAuthViewTest(_BaseMFAViewTest):
    url = reverse("multi_factor_auth-detail")

    def test_get(self):
        # Check for HTTP401.
        self.check_get_response(self.url, status.HTTP_403_FORBIDDEN)

        # Check for HTTP200 - MFA for a disabled user.
        resp = self.check_get_response(
            self.url, status.HTTP_200_OK, user=self.user1
        )
        resp_json = resp.data
        self.assertEqual(resp_json["enabled"], False)
        with self.assertRaises(KeyError):
            resp_json["challenge_type"]

        # Check for HTTP200 - MFA Enabled User.
        resp = self.check_get_response(
            self.url, status.HTTP_200_OK, user=self.user2
        )
        resp_json = resp.data
        self.assertEqual(resp_json["enabled"], True)
        self.assertEqual(resp_json["challenge_type"], SMS)

    def test_delete(self):
        # Check for HTTP401.
        self.check_delete_response(self.url, status.HTTP_403_FORBIDDEN)

        # Check for HTTP400 - MFA for a disabled user.
        resp = self.check_delete_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1)
        self.assertEqual(resp.data, {
            "detail": strings.DISABLED_ERROR
        })

        # Check for HTTP200.
        self.mfa_2.refresh_backup_code()
        self.check_delete_response(
            self.url, status.HTTP_204_NO_CONTENT, user=self.user2)
        instance = mfa_settings.MFA_MODEL.objects.get(user=self.user2)
        self.assertFalse(instance.enabled)
        self.assertEqual(instance.challenge_type, DISABLED)
        self.assertEqual(instance.backup_code, "")
        self.assertEqual(instance.phone_number, "")


class SMSChallengeRequestViewTest(_BaseMFAViewTest):
    url = reverse("sms_request-detail")

    def test_unauthorized(self):
        self.check_put_response(self.url, status.HTTP_403_FORBIDDEN)

    def test_already_enabled(self):
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user2,
            data={"phone_number": self.phone_number})
        self.assertEqual(resp.data, {"detail": [strings.ENABLED_ERROR]})

    def test_bad_phone_numbers(self):
        # No phone number inputted.
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1)
        self.assertEqual(resp.data, {
            "phone_number": ["This field is required."]
        })

        # Invalid phone number.
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1,
            data={"phone_number": "bad_number"})
        self.assertEqual(resp.data, {
            "phone_number": [strings.INVALID_PHONE_NUMBER_ERROR]
        })

    @patch("deux.serializers.MultiFactorChallenge")
    def test_failed_sms_error(self, challenge):
        challenge.return_value.generate_challenge.side_effect = (
            FailedChallengeError("Error Message."))
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1,
            data={"phone_number": self.phone_number})
        self.assertEqual(resp.data, {
            "detail": "Error Message."
        })

    @patch("deux.serializers.MultiFactorChallenge")
    def test_success(self, challenge):
        resp = self.check_put_response(
            self.url, status.HTTP_200_OK, user=self.user1,
            data={"phone_number": self.phone_number})
        self.assertEqual(resp.data, {
            "enabled": False, "phone_number": self.phone_number
        })
        challenge.assert_called_once_with(self.mfa_1, SMS)
        challenge.return_value.generate_challenge.assert_called_once_with()


class SMSChallengeVerifyViewTest(_BaseMFAViewTest):
    url = reverse("sms_verify-detail")

    def test_unauthorized(self):
        self.check_put_response(self.url, status.HTTP_403_FORBIDDEN)

    def test_already_enabled(self):
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user2,
            data={"mfa_code": "code"})
        self.assertEqual(resp.data, {"detail": [strings.ENABLED_ERROR]})

    def test_incorrect_mfa_codes(self):
        # Check for failure with incorrect mfa_code.
        resp = self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1,
            data={"mfa_code": "bad_code"}
        )
        self.assertEqual(resp.data, {
            "mfa_code": [strings.INVALID_MFA_CODE_ERROR]
        })

        # Check for failure with None mfa_code.
        self.check_put_response(
            self.url, status.HTTP_400_BAD_REQUEST,
            user=self.user1, data=None
        )
        self.assertEqual(resp.data, {
            "mfa_code": [strings.INVALID_MFA_CODE_ERROR]
        })

    def test_success(self):
        mfa_code = generate_mfa_code(self.mfa_1.sms_bin_key)
        resp = self.check_put_response(
            self.url, status.HTTP_200_OK, user=self.user1,
            data={"mfa_code": mfa_code}
        )
        resp_json = resp.data
        self.assertEqual(resp_json["enabled"], True)
        self.assertEqual(resp_json["challenge_type"], SMS)
        instance = mfa_settings.MFA_MODEL.objects.get(user=self.user1)
        self.assertTrue(instance.enabled)
        self.assertEqual(instance.challenge_type, SMS)


class BackupCodesViewTest(_BaseMFAViewTest):
    url = reverse("backup_code-detail")

    def test_get(self):
        # Check for HTTP401.
        self.check_get_response(self.url, status.HTTP_403_FORBIDDEN)

        # Check for HTTP400 - MFA for a disabled user.
        resp = self.check_get_response(
            self.url, status.HTTP_400_BAD_REQUEST, user=self.user1)
        self.assertEqual(resp.data, {
            "backup_code": strings.DISABLED_ERROR
        })

        # Check for HTTP200.
        resp = self.check_get_response(
            self.url, status.HTTP_200_OK, user=self.user2)
        self.assertEqual(
            len(resp.data["backup_code"]), mfa_settings.BACKUP_CODE_DIGITS)



from __future__ import absolute_import, unicode_literals

#: Represents the ``DISABLED`` state of MFA.
DISABLED = ""

#: Represents the state of using ``SMS`` for MFA.
SMS = "sms"

#: A tuple of all support challenge types.
CHALLENGE_TYPES = (SMS,)



"""testproj URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.9/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.conf.urls import include, url

urlpatterns = [
    url(r"^api-auth/",
        include("rest_framework.urls", namespace="rest_framework")),
    url(r"^mfa/", include("deux.urls")),
    url(r"^mfa/authtoken/",
        include("deux.authtoken.urls", namespace="authtoken"),
    ),
    url(r"^mfa/oauth2/",
        include("deux.oauth2.urls", namespace="oauth2"),
    ),
]



from __future__ import absolute_import, unicode_literals

from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from deux import views

urlpatterns = [
    url(r"^$", views.MultiFactorAuthDetail.as_view(),
        name="multi_factor_auth-detail"),
    url(r"^sms/request/$", views.SMSChallengeRequestDetail.as_view(),
        name="sms_request-detail"),
    url(r"^sms/verify/$", views.SMSChallengeVerifyDetail.as_view(),
        name="sms_verify-detail"),
    url(r"^recovery/$", views.BackupCodeDetail.as_view(),
        name="backup_code-detail"),
]

urlpatterns = format_suffix_patterns(urlpatterns)



from __future__ import absolute_import, unicode_literals

from deux.abstract_models import AbstractMultiFactorAuth


class MultiFactorAuth(AbstractMultiFactorAuth):
    """
    class::MultiFactorAuth()

    Blank extension of ``AbstractMultiFactorAuth`` that is used as the
    default model in this package.
    """
    pass



from __future__ import absolute_import, unicode_literals

import json
import six
from base64 import b64encode
from mock import patch
from oauth2_provider.models import get_application_model

from django.core.urlresolvers import reverse
from rest_framework import status

from deux.app_settings import mfa_settings
from deux.constants import SMS
from deux.services import generate_mfa_code
from deux.tests.test_base import BaseUserTestCase

Application = get_application_model()


class MFAOAuth2TokenTests(BaseUserTestCase):
    url = reverse("oauth2:token")

    def setUp(self):
        self.simpleUserSetup()
        self.application = Application.objects.create(
            name="Test Password Application",
            user=self.user1,
            authorization_grant_type=Application.GRANT_PASSWORD,
        )
        self.headers = self._get_basic_auth_header(
            self.application.client_id, self.application.client_secret)

        self.mfa = mfa_settings.MFA_MODEL.objects.create(user=self.user2)
        self.mfa.enable(SMS)
        self.phone_number = "1234567890"
        self.mfa.phone_number = self.phone_number
        self.mfa.save()
        self.backup_code = self.mfa.refresh_backup_code()
        self.mfa_code = generate_mfa_code(self.mfa.sms_bin_key)

    def test_incorrect_credentials(self):
        data = self._get_data(
            username=self.user1.username, password="wrong password")
        response = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data,
            headers=self.headers)
        self._assert_error_msg(
            response, "Unable to log in with provided credentials.")

    def test_inactive_user(self):
        self.user1.is_active = False
        self.user1.save()
        data = self._get_data(
            username=self.user1.username, password=self.user1.password)
        response = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data,
            headers=self.headers)
        self._assert_error_msg(
            response, "Unable to log in with provided credentials.")

    def test_get_token_mfa_object_does_not_exist(self):
        data = self._get_data(
            username=self.user1.username, password=self.password1)
        response = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data, headers=self.headers)
        self._assert_authenticated(response)

    def test_get_token_mfa_not_required(self):
        self.mfa.disable()
        data = self._get_data()
        response = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data, headers=self.headers)
        self._assert_authenticated(response)

    @patch("deux.oauth2.validators.MultiFactorChallenge")
    def test_get_token_mfa_required(self, challenge):
        data = self._get_data()
        response = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data, headers=self.headers)
        content = json.loads(response.content.decode("utf-8"))
        self.assertTrue(content.get("mfa_required"))
        self.assertEqual(content.get("mfa_type"), SMS)
        challenge.assert_called_once_with(self.mfa, SMS)

    def test_login_success_with_mfa_code(self):
        data = self._get_data(mfa_code=self.mfa_code)
        response = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data, headers=self.headers)
        self._assert_authenticated(response)

    def test_login_fail_with_invalid_mfa_code(self):
        bad_code = six.text_type(int(self.mfa_code) + 1)
        data = self._get_data(mfa_code=bad_code)
        response = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data,
            headers=self.headers)
        self._assert_error_msg(response, "Please enter a valid code.")

    def test_login_success_with_backup_code(self):
        data = self._get_data(backup_code=self.backup_code)
        response = self.check_post_response(
            self.url, status.HTTP_200_OK, data=data, headers=self.headers)
        self._assert_authenticated(response)

    def test_login_fail_with_invalid_backup_code(self):
        bad_backup_code = "abcdef123456"
        data = self._get_data(backup_code=bad_backup_code)
        response = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data,
            headers=self.headers)
        self._assert_error_msg(response, "Please enter a valid backup code.")

    def test_login_fail_with_both_codes(self):
        data = self._get_data(
            mfa_code=self.mfa_code, backup_code=self.backup_code)
        response = self.check_post_response(
            self.url, status.HTTP_400_BAD_REQUEST, data=data,
            headers=self.headers)
        self._assert_error_msg(
            response,
            "Login does not take both a verification and backup code."
        )

    def _assert_authenticated(self, response):
        content = json.loads(response.content.decode("utf-8"))
        self.assertIsNotNone(content.get("access_token"))
        self.assertIsNotNone(content.get("refresh_token"))

    def _assert_error_msg(self, response, msg):
        content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(content["detail"], msg)

    def _get_data(
            self, username=None, password=None, mfa_code=None,
            backup_code=None):
        return {
            "grant_type": "password",
            "username": username or self.user2.username,
            "password": password or self.password2,
            "mfa_code": mfa_code,
            "backup_code": backup_code
        }

    def _get_basic_auth_header(self, client_id, client_secret):
        """
        Return a dict containg the correct headers to set to make HTTP
        an Auth request to Oauth2.
        """
        id_secret = '{id}:{secret}'.format(id=client_id, secret=client_secret)
        auth_string = b64encode(id_secret.encode('utf-8'))
        return {'HTTP_AUTHORIZATION': 'Basic ' + auth_string.decode("utf-8")}



from __future__ import absolute_import, unicode_literals

import six
from oauthlib.oauth2 import OAuth2Error

from rest_framework import status


class InvalidLoginError(OAuth2Error):
    """
    Generic exception for a failed login attempt through OAuth2. This exception
    will result in a 400 Bad Request error in the OAuth API.
    """

    def __init__(self, message):
        """
        Initializes this error with the given error message.

        :param message: The error message describing this exception.
        """
        super(Exception, self).__init__(message)

    @property
    def twotuples(self):
        """
        Returns a list of tuples that will be converted to the error response.
        This method override the ``two_tuples`` method from ``OAuth2Error``.
        """
        return [("detail", six.text_type(self))]


class ChallengeRequiredMessage(OAuth2Error):
    """
    This exception is used to prompt the user for an MFA code.

    This exception is used when a user passes in their username and password,
    and they have MFA enabled.
    """

    #: This exception returns a 200 response.
    status_code = status.HTTP_200_OK

    def __init__(self, challenge_type):
        """
        Initalizes this exception class with a challenge type.

        :param challenge_type: The challenge type the user should expect.
        """
        super(Exception, self).__init__(challenge_type)

    @property
    def twotuples(self):
        """
        Returns a list of tuples that will be converted to the error response.
        This method override the ``two_tuples`` method from ``OAuth2Error``.
        """
        return [
            ("mfa_required", True),
            ("mfa_type", six.text_type(self)),
        ]



from __future__ import absolute_import, unicode_literals

import six
from binascii import unhexlify
from mock import patch

from django.test import TestCase
from django_otp.util import random_hex

from deux.app_settings import mfa_settings
from deux.constants import SMS
from deux.services import (
    MultiFactorChallenge,
    generate_mfa_code,
    verify_mfa_code,
)

from .test_base import BaseUserTestCase


class GenerateMFACodeTests(TestCase):

    def setUp(self):
        self.bin_key = unhexlify(random_hex())

    def test_generate_mfa_code(self):
        mfa_code = generate_mfa_code(self.bin_key)
        self.assertEqual(len(mfa_code), mfa_settings.MFA_CODE_NUM_DIGITS)

    def test_time_based_mfa_code(self):
        mfa_code_0 = generate_mfa_code(self.bin_key, drift=0)
        mfa_code_1 = generate_mfa_code(self.bin_key, drift=1)
        self.assertNotEquals(mfa_code_0, mfa_code_1)


class VerifyMFACodeTests(TestCase):

    def setUp(self):
        self.bin_key = unhexlify(random_hex())

    def test_verify_mfa_code_success(self):
        mfa_code_tests = (
            generate_mfa_code(self.bin_key, -1),
            generate_mfa_code(self.bin_key, 0),
            generate_mfa_code(self.bin_key, 1)
        )
        for mfa_code in mfa_code_tests:
            self.assertTrue(verify_mfa_code(self.bin_key, mfa_code))

    def test_verify_mfa_code_fail(self):
        int_mfa_code = int(generate_mfa_code(self.bin_key, 0))
        mfa_code_tests = (
            None,
            "",
            generate_mfa_code(self.bin_key, -3),
            generate_mfa_code(self.bin_key, -2),
            generate_mfa_code(self.bin_key, 2),
            generate_mfa_code(self.bin_key, 3),
            six.text_type(int_mfa_code + 1).zfill(
                mfa_settings.MFA_CODE_NUM_DIGITS),
            "abcdef"
        )
        for mfa_code in mfa_code_tests:
            self.assertFalse(verify_mfa_code(self.bin_key, mfa_code))


class MultiFactorChallengeTests(BaseUserTestCase):

    def setUp(self):
        self.simpleUserSetup()
        self.mfa = mfa_settings.MFA_MODEL.objects.create(user=self.user1)

    @patch("deux.services.mfa_settings.SEND_MFA_TEXT_FUNC")
    @patch("deux.services.generate_mfa_code")
    def test_sms_challenge(self, generate_mfa_code, text_function):
        generate_mfa_code.return_value = "123456"
        MultiFactorChallenge(self.mfa, SMS).generate_challenge()
        text_function.assert_called_once_with(
            mfa_instance=self.mfa,
            mfa_code="123456")

    def test_invalid_challenge(self):
        fail_tests = ("SMS", "abc", 123)
        for test in fail_tests:
            with self.assertRaises(AssertionError):
                MultiFactorChallenge(self.mfa, test)



from __future__ import absolute_import, unicode_literals

from oauth2_provider.views import TokenView

from deux.oauth2.backends import MFARequestBackend
from deux.oauth2.validators import MFAOAuth2Validator


class MFATokenView(TokenView):
    """
    class::MFATokenView()

    Extends OAuth's base TokenView to support MFA.
    """

    #: Use Deux's custom backend for the MFA OAuth api.
    oauthlib_backend_class = MFARequestBackend

    #: Use Deux's custom validator for the MFA OAuth api.
    validator_class = MFAOAuth2Validator



from __future__ import absolute_import, unicode_literals

from django.core.validators import RegexValidator

from deux import strings

#: Regex validator for phone numbers.
phone_number_validator = RegexValidator(
    regex=r"^(\d{7,15})$",
    message=strings.INVALID_PHONE_NUMBER_ERROR)
